CONFIG_FILEPATH = ''
ANCHOR_NAMES = {'splitStemUpSE',
                'splitStemUpSW',
                'splitStemDownNE',
                'splitStemDownNW',
                'stemUpSE',
                'stemDownNW',
                'stemUpNW',
                'stemDownSW',
                'nominalWidth',
                'numeralTop',
                'numeralBottom',
                'cutOutNE',
                'cutOutSE',
                'cutOutSW',
                'cutOutNW',
                'graceNoteSlashSW',
                'graceNoteSlashNE',
                'graceNoteSlashNW',
                'graceNoteSlashSE',
                'repeatOffset',
                'noteheadOrigin',
                'opticalCenter'}

ENGRAVING_DEFAULT_KEYS = {'arrowShaftThickness',
                          'barlineSeparation',
                          'beamSpacing',
                          'beamThickness',
                          'bracketThickness',
                          'dashedBarlineDashLength',
                          'dashedBarlineGapLength',
                          'dashedBarlineThickness',
                          'hairpinThickness',
                          'hBarThickness',
                          'legerLineExtension',
                          'legerLineThickness',
                          'lyricLineThickness',
                          'octaveLineThickness',
                          'pedalLineThickness',
                          'repeatBarlineDotSeparation',
                          'repeatEndingLineThickness',
                          'slurEndpointThickness',
                          'slurMidpointThickness',
                          'staffLineThickness',
                          'stemThickness',
                          'subBracketThickness',
                          'textFontFamily',
                          'textEnclosureThickness',
                          'thickBarlineThickness',
                          'thinBarlineThickness',
                          'thinThickBarlineSeparation',
                          'tieEndpointThickness',
                          'tieMidpointThickness',
                          'tupletBracketThickness'}
